/**
 ** OrthoAltitudeMatch - Represent elevation maps matching for pair of scenes 
 **
 ** Copyright (C) 2024 Elphel, Inc.
 **
 ** -----------------------------------------------------------------------------**
 **
 **  PairwiseOrthoMatch.java is free software: you can redistribute it and/or modify
 **  it under the terms of the GNU General Public License as published by
 **  the Free Software Foundation, either version 3 of the License, or
 **  (at your option) any later version.
 **
 **  This program is distributed in the hope that it will be useful,
 **  but WITHOUT ANY WARRANTY; without even the implied warranty of
 **  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 **  GNU General Public License for more details.
 **
 **  You should have received a copy of the GNU General Public License
 **  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 ** -----------------------------------------------------------------------------**
 **
 */

package com.elphel.imagej.orthomosaic;

import java.awt.Point;
import java.awt.Rectangle;
import java.io.IOException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Calendar;

import com.elphel.imagej.calibration.CalibrationFileManagement;
import com.elphel.imagej.cameras.CLTParameters;

import ij.IJ;

public class OrthoAltitudeMatch {

	public static boolean altutudeMatchPairs(
			CLTParameters       clt_parameters,
			OrthoMapsCollection orthoMapsCollection, 
			int [][]            available_pairs, 
			boolean             alt_overwrite, // overwrite existing altitude match pairs
			boolean             alt_pairwise,  // use pairwise affines if available
			double              alt_sigma,     // 5.0; Reduce weight of the border tiles, Gaussian sigma in tiles to apply to weights.
			double              alt_abs_outliers, //  =   3.0;     // remove absolute outliers when fitting planes 
			double              alt_outliers,  // 0.05; Remove outliers when fitting planes, removed fraction.
			int                 alt_refine,    // 1; Refine altitude difference plane after removing outliers (0 - no outlier removal, 1 - remove outliers and refine once, ...)
			double              metric_error,
			// log/save parameters
			boolean             save_each,
			boolean             log_append,
			String              log_path,
			String              orthoMapsCollection_path,
			int                 debugLevel) {
		int [] indices = orthoMapsCollection.getScenesFromPairs( // may be shorter, each element - absolute scene number used in pairs
				available_pairs,  // pairs_defined_abs,// int [][] pairs,
				null);            // int [] indices_in)  // preselected indices or null
		boolean updateStatus = true;
		int [][] condensed_pairs = orthoMapsCollection.condensePairs (available_pairs, indices);
		int alt_zoom_offs = -3;   // altitude resolution is 3 steps lower than the images
		if (indices.length < 2) {
			System.out.println("too few scenes remain: "+indices.length);
			return false;
		}
		OrthoMap[] ortho_maps = orthoMapsCollection.getMaps();
		int min_zoom_all = ortho_maps[indices[0]].getOriginalZoomLevel();
		int max_zoom_all = ortho_maps[indices[0]].getOriginalZoomLevel();
		for (int i = 1; i < indices.length; i++) {
			min_zoom_all = Math.min(min_zoom_all, ortho_maps[indices[i]].getOriginalZoomLevel());
			max_zoom_all = Math.max(max_zoom_all, ortho_maps[indices[i]].getOriginalZoomLevel());
		}
		int alt_zoom_lev = Math.min(max_zoom_all+alt_zoom_offs, min_zoom_all);
		double pix_size_meters = OrthoMap.getPixelSizeMeters(alt_zoom_lev);
		if (debugLevel > 0 ) {
			System.out.println("max_zoom="+max_zoom_all+ " min_zoom="+min_zoom_all+" alt_zoom_lev="+alt_zoom_lev);
		}
		int min_scene =      0; // use if needs to continue (during development)
		if (log_append && (log_path != null)) { // assuming directory exists
			StringBuffer sb = new StringBuffer();
			sb.append(new SimpleDateFormat("yyyy/MM/dd HH:mm:ss").format(Calendar.getInstance().getTime())+"\n");
			
			sb.append("number of scenes pairs\t"+     available_pairs.length+"\n");
			sb.append("number of scenes\t"+     indices.length+"\n");
			sb.append("minimal zoom level\t"+   min_zoom_all+"\n");
			sb.append("maximal zoom level\t"+   max_zoom_all+"\n");
			sb.append("altitude zoom level\t"+  alt_zoom_lev+"\n");
			
//			sb.append(String.format("%4s\t%4s\t%17s\t%17s\t%6s\t%3s\t%6s\t%6s\t%7s\n",
//					"scn1","scn2","timestamp1","timestamp2","ovrlp","zl","RMS-sp","RMSfin","fzl","removed"));
			sb.append(String.format("%4s\t%4s\t%17s\t%17s\t%6s\t%3s\t%6s\t%7s\t%7s\t%7s\n",
					"scn1","scn2","timestamp1","timestamp2","ovrlp","zl","RMS","tiltX","tiltY","offs"));
			CalibrationFileManagement.saveStringToFile (
					log_path,          //String path,
					sb.toString(), // data,
					true); // boolean append)
			if (debugLevel>-3) {
				System.out.print(sb.toString());
			}
		}
		// create altitude map
		int [] wh =     new int[2];
		int [] origin = new int[2];
		boolean show_centers = true;
		boolean  bounds_to_indices = true;
		double [][] centers = show_centers? (new double [indices.length][]): null;
		double [][] alt_multi = orthoMapsCollection.renderMultiDouble (
				new double [indices.length][] , // double [][]   ground_planes, // null - images, non-null altitudes. use new double[2][3] for old way alt
				indices,                        // int []        indices, // null or which indices to use (normally just 2 for pairwise comparison)
				bounds_to_indices,              // boolean       bounds_to_indices,
				null,                           // affines,    // double [][][] affines, // null or [indices.length][2][3]
				null,                           // double [][]   equalize,
				true,                           // boolean       ignore_equalize,
				null,                           // warp,       // FineXYCorr    warp,,
				alt_zoom_lev,                       // int         zoom_level,
				wh,                             // int []      wh,
				origin,                         // int []      origin){ // maps[0] as a reference
				centers);                       // double [][] centers)
		final int width = wh[0];
//		final int height = wh[1];
		
//		int num_pairs = 0; // available_pairs.length
//		ArrayList<Point> failed_pairs = new ArrayList<Point>();
		for (int npair = 0; npair < condensed_pairs.length; npair++) {
			int [] cpair = condensed_pairs[npair]; // index alt_multi
			int [] ipair = {indices[cpair[0]], indices[cpair[1]]};
			if (updateStatus) {
				IJ.showStatus("Processing pair "+npair+" of "+condensed_pairs.length+" "+ipair[0]+"->"+ipair[1]);
			}
			/*
			
			PairwiseOrthoMatch pairwiseOrthoMatch = ortho_maps[ipair[0]].getMatch(ortho_maps[ipair[1]].getName(), true).clone(); // ?
			double [][] daffine = null;
    		if (pairwiseOrthoMatch != null) {
    			double [] enuOffset = ortho_maps[ipair[1]].enuOffsetTo(ortho_maps[ipair[0]]);
    			double [] rd = {enuOffset[0], -enuOffset[1]}; // {right,down} of the image
    			daffine = pairwiseOrthoMatch.getAffine();
    			if ((daffine != null) && alt_pairwise) {
    			} else { // combine differential affine from individual
    				double [][] aff0 = ortho_maps[ipair[0]].getAffine();            			
    				double [][] aff1 = ortho_maps[ipair[1]].getAffine();
    				PairwiseOrthoMatch aff_match = new PairwiseOrthoMatch (
    						aff0, // double [][] affine0,
    						aff1, // double [][] affine1,
    						rd);  // double [] rd);
    				daffine = aff_match.getAffine();
    				pairwiseOrthoMatch.setAffine(daffine);
    			}
    		} else {
    			System.out.println("BUG: Missing pair for ["+ipair[0]+", "+ipair[1]+"] ");
    			continue;
    		}
    		*/
			PairwiseOrthoMatch pairwiseOrthoMatch = ortho_maps[ipair[0]].getMatch(ortho_maps[ipair[1]].getName(), true); // ?
    		if (pairwiseOrthoMatch == null) {
    			System.out.println("BUG: Missing pair for ["+ipair[0]+", "+ipair[1]+"] ");
    			continue;
    		}
			if (ipair[0] < min_scene) {
				System.out.println ("Skipping "+ipair[0]+":"+ipair[1]+" until "+min_scene);
				continue;
			}
			if ((pairwiseOrthoMatch.getAltData() != null) && !alt_overwrite)  {
				System.out.println ("Skipping "+ipair[0]+":"+ipair[1]+" as it has alt_data defined and alt_overwrite == false .");
				continue;
			}
			// pairwiseOrthoMatch is clone() - not anymore
			
	    	int min_zoom_lev = ortho_maps[ipair[0]].getOriginalZoomLevel();
	    	int max_zoom_lev = ortho_maps[ipair[0]].getOriginalZoomLevel();
	    	double max_agl =   ortho_maps[ipair[0]].getAGL();
	    	for (int i = 0; i < ipair.length; i++) {
	    		max_agl =      Math.max(max_agl, ortho_maps[ipair[i]].getAGL());	    		
	    		min_zoom_lev = Math.min(min_zoom_lev, ortho_maps[ipair[i]].getOriginalZoomLevel());
	    		max_zoom_lev = Math.max(max_zoom_lev, ortho_maps[ipair[i]].getOriginalZoomLevel());
	    	}
	    	double [][] alt_slices = {alt_multi[cpair[0]], alt_multi[cpair[1]]};
	    	Rectangle woi_overlap = OrthoMap.getDefinedBounds(
	    			alt_slices, // final double [][] data_slices,
	    			width); // final int width);
	    	double [] diff_data = OrthoMap. subtractWoi(
	    			alt_multi[cpair[0]], // final double [] data0,
	    			alt_multi[cpair[1]], // final double [] data1,
	    			width,               // final int       width,
	    			woi_overlap);        // Rectangle woi_in);
	    	
	        double []  xy0 = {0.5*woi_overlap.width, 0.5*woi_overlap.height};
	        double [] weight = null; //  make it fading?
	        if (alt_sigma > 0) {
	        	weight = OrthoMap.getBorderWeights(
	        			diff_data, //final double []  data,
	        			alt_sigma, // final double     sigma,
	        			woi_overlap.width);    // int              width)
	        }
	        
	        boolean [] mask = null;
	        double [] alt_data5 = null;
        	int    num_bins = 1000;
    		for (int ntry = 0; ntry <= alt_refine; ntry++) {
    			alt_data5 = OrthoMap.getPlane(
    	        		diff_data,         // final double []   data,
    	    			mask,              // final boolean [] mask,
    	    			weight,            // final double []  weight,
    	    			woi_overlap.width, // final int        width,
    	    			xy0);              // final double []  xy0) {
    	        if ((alt_outliers > 0) && (ntry < alt_refine)){ // not the last pass
	    			mask = OrthoMap.removeRelativeLowHigh (
	    					diff_data, //  final double [] data,
	    					null, // mask, // final boolean [] mask_in, // new mask for all data and latest plane
	    					alt_abs_outliers, // final double abs_diff,
	    					alt_outliers,     // final double rel_frac,
	    					alt_data5,        // final double []  ground_plane, // tiltx,tilty, offs, x0(pix), y0(pix) or null
	    					woi_overlap.width,            // final int     width, // only used with ground_plane != null;
	    					num_bins);        // final int    num_bins)	    			
    	        } else {
    	        	break;
    	        }
    		}	        
	        
	        
//	        double [] alt_data = new double[3];
//	        System.arraycopy(alt_data5, 0, alt_data, 0, alt_data.length);
	        double [] alt_data = {alt_data5[0]/pix_size_meters, alt_data5[1]/pix_size_meters,alt_data5[2]};
	        pairwiseOrthoMatch.setAltData(alt_data);
	        
			if (log_append && (log_path != null)) { // assuming directory exists
				StringBuffer sb = new StringBuffer();
				sb.append(String.format("%4d\t%4d\t%s\t%s\t%6.4f\t%3d\t%6.4f\t%7.4f\t%7.4f\t%7.3f\n",
						ipair[0], ipair[1], ortho_maps[ipair[0]].getName(), ortho_maps[ipair[1]].getName(),
						pairwiseOrthoMatch.overlap, pairwiseOrthoMatch.zoom_lev, pairwiseOrthoMatch.rms,
						pairwiseOrthoMatch.alt_data[0],pairwiseOrthoMatch.alt_data[1],pairwiseOrthoMatch.alt_data[2]));
				CalibrationFileManagement.saveStringToFile (
						log_path,          //String path,
						sb.toString(), // data,
						true); // boolean append)
			}
	        
	        
	        
	 //pix_size_meters   	
	    	
	    	/*
        	double agl_ratio = max_agl/50.0;
        	double metric_error_adj = metric_error * agl_ratio * agl_ratio; // metric_error settings is good for 50m. Increase for higher Maybe squared?
	    	int initial_zoom = max_zoom_lev - 4;  // another algorithm?
	    	// overlaps
	    	double overlap_frac = pairwiseOrthoMatch.getOverlap(); // .pairsGraph.getOverlap(next_pair);
			// unityAffine()
	    	// use unityAffine() for 0; getaffine for second?
			double [][] affine0 = OrthoMapsCollection.unityAffine(); // ortho_maps[ipair[0]].getAffine();
			double [][] affine1 = daffine; // ortho_maps[ipair[1]].getAffine();
			double [][][] affines = new double[][][] {affine0,affine1};
			boolean success = true;
			Point pair = new Point(ipair[0],ipair[1]);
			double lores_rms = pairwiseOrthoMatch.getRMS();
			// high-res

			affines[1][0] = affines[1][0].clone(); 
			affines[1][1] = affines[1][1].clone();
			Rectangle woi = new Rectangle(); // used to return actual woi from correlateOrthoPair()		
			correlateOrthoPair(
					clt_parameters,  // CLTParameters    clt_parameters,
					pairwiseOrthoMatch, //PairwiseOrthoMatch pairwiseOrthoMatch, // will return statistics
					0,                // 			int              min_overlap,
					max_std,         // double           max_std,      // maximal standard deviation to limit center area  
					min_std_rad,     // double           min_std_rad,  // minimal radius of the central area (if less - fail)
					frac_remove,     // double           frac_remove, //  =        0.25
					metric_error_adj,// 			double           metric_error,
					ignore_prev_rms, // boolean ignore_prev_rms,
					num_tries,       //  = 5int              num_tries, //  = 5
					false, // ,            // boolean          calc_warp, (will return null if false)
					batch_mode,      // boolean          batch_mode,
					ipair,           // String []        gpu_spair,
					affines,         // double [][][]    affines, // on top of GPS offsets
					woi,             // Rectangle        woi,
					min_zoom_lev,    // int              zoom_lev,
					false,           // show_vf,         // boolean show_vf,
					null,            // ground_planes,   // double [][]      ground_planes, // null or double[2] - will return ground planes
					rad_fraction,    // double           rad_fraction,
					max_tile_rad,    // double           max_tile_rad, //  = 30;
					fill_fraction,   // double           fill_fraction,
					fill_fraction_final, // double           fill_fraction_final,
					ease_nosfm,      // double           ease_nosfm,
					null,            //  double []        max_rms_iter, //  = {1.0, 0.6};//
					pull_skew,       // double           pull_skew,        // ~rotation, = 0 fraction of the total weight == 1
					pull_tilt,       // double           pull_tilt,        // > 0
					pull_scale,      // double           pull_scale,       // = 0
					debugLevel-4);   // final int        debugLevel)
	*/		
		}
		if (orthoMapsCollection_path != null) {
        	try {
        		orthoMapsCollection.writeOrthoMapsCollection(orthoMapsCollection_path);
        	} catch (IOException e) {
        		// TODO Auto-generated catch block
        		e.printStackTrace();
        	}
        	if (debugLevel > -4) {
        		System.out.println("Saved data to "+ orthoMapsCollection_path);
        	}
		}
		return true;
    }
	
	
	
}
